@@ -1,6 +1,6 @@
 // -*-mode:c++; c-style:k&r; c-basic-offset:4;-*-
 //
-// Copyright 2011-2021, Julian Catchen <jcatchen@illinois.edu>
+// Copyright 2011-2024, Julian Catchen <jcatchen@illinois.edu>
 //
 // This file is part of Stacks.
 //
@@ -107,11 +107,16 @@ parse_input_record(Seq *s, RawRead *r)
     uint  lim;
     //
     // Count the number of colons to differentiate Illumina version.
-    // CASAVA 1.8+ has a FASTQ header like this:
+    //
+    // Illumina's BCL Convert program or CASAVA 1.8+ has a FASTQ header like this:
     //  @HWI-ST0747:155:C01WHABXX:8:1101:6455:26332 1:N:0:
+    // 
     // Or, with the embedded barcode:
     //  @HWI-ST1233:67:D12GNACXX:7:2307:14604:78978 1:N:0:ATCACG
-    //
+    // 
+    // Or, with a UMI (unique molecular identifier -- for PCR duplicate control):
+    //  @LH00346:7:22CH5TLT3:5:1101:1407:1032:rTACCCAGT 1:N:0:ACATTGCG
+    //   -the 'r' in the UMI field indicates it is reverse complemented.
     //
     // Or, parse FASTQ header from previous versions that looks like this:
     //  @HWI-ST0747_0141:4:1101:1240:2199#0/1
@@ -126,16 +131,16 @@ parse_input_record(Seq *s, RawRead *r)
         hash_cnt  += *q == '#' ? 1 : 0;
     }
 
-    if (colon_cnt == 9 && hash_cnt == 0) {
+    if ((colon_cnt >= 9 && colon_cnt <= 10) && hash_cnt == 0) {
         r->fastq_type = illv2_fastq;
         //
-        // According to Illumina manual, "CASAVA v1.8 User Guide" page 41:
+        // Colon_cnt == 9; According to Illumina manual, "CASAVA v1.8 User Guide" page 41:
         // @<instrument>:<run number>:<flowcell ID>:<lane>:<tile>:<x-pos>:<y-pos> <read>:<is filtered>:<control number>:<index sequence>
         //
         for (p = s->id, q = p; *q != ':' && q < stop; q++);
         if (q < stop) {
             *q = '\0';
-            strcpy(r->machine, p);
+            r->machine(p);
             *q = ':';
         }
 
@@ -172,12 +177,28 @@ parse_input_record(Seq *s, RawRead *r)
             *q = ':';
         }
 
-        for (p = q+1, q = p; *q != ' ' && q < stop; q++);
-        if (q < stop) {
-            *q = '\0';
-            r->y = atoi(p);
-            *q = ' ';
-        }
+	for (p = q+1, q = p; *q != ' ' && q < stop; q++);
+	if (q < stop) {
+	    *q = '\0';
+	    r->y = atoi(p);
+	}
+
+	if (colon_cnt == 10) {
+	    //
+	    // According to Illumina's BCL Convert program an extra field occurs containing the UMI:
+	    // @<instrument>:<run number>:<flowcell ID>:<lane>:<tile>:<x-pos>:<y-pos>:<UMI> <read>:<is filtered>:<control number>:<index sequence>
+	    //
+	    *q = ';';
+	    r->fastq_type = illv3_fastq;
+
+	    for (p = q + 1, q = p; *q != ':' && q < stop; q++);
+	    if (q < stop) {
+		*q = '\0';
+		strncpy(r->umi, p, id_len);
+	    }
+	}
+	if (q < stop)
+	    *q = ' ';
 
         for (p = q+1, q = p; *q != ':' && q < stop; q++);
         if (q < stop) {
@@ -270,7 +291,7 @@ parse_input_record(Seq *s, RawRead *r)
         for (p = s->id, q = p; *q != ':' && q < stop; q++);
         if (q < stop) {
             *q = '\0';
-            strcpy(r->machine, p);
+            r->machine(p);
             *q = ':';
         }
 
@@ -315,8 +336,16 @@ parse_input_record(Seq *s, RawRead *r)
     } else {
         r->fastq_type = generic_fastq;
 
-        strncpy(r->machine, s->id, id_len);
-        r->machine[id_len] = '\0';
+        //
+        // Check if the header already contains a "/1" or "/2" notation at the end,
+        // from previous parsing of the file (e.g. Sequence Read Archve data). Remove
+        // if present.
+        //
+        char *p = s->id + strlen(s->id);
+        if (*(p - 2) == '/' && (*(p - 1) == '1' || *(p - 1) == '2'))
+            *(p - 2) = '\0';
+
+        r->machine(s->id);
     }
 
     uint len = strlen(s->seq);
@@ -451,6 +480,14 @@ check_quality_scores(RawRead *href, int qual_offset, int score_limit, int len_li
     //   |         |         |         |         |
     //  64        74        84        94       104
     //   0        10(90%)   20(99%)   30(99.9%) 40(99.99%)
+    //
+    // Illumina 1.8+ encodes phred scores between ASCII values 33 (0 quality) and 74 (41 quality)
+    //
+    //   !"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJK
+    //   |         |         |         |          |
+    //  33        43        53        63         73
+    //   0        10(90%)   20(99%)   30(99.9%)  40(99.99%)
+
     //
     //
     // Drop sequence if the average phred quality score drops below a threshold within a sliding window.
@@ -459,12 +496,9 @@ check_quality_scores(RawRead *href, int qual_offset, int score_limit, int len_li
     //
     // Convert the encoded quality scores to their integer values
     //
-    // cerr << "Integer scores: ";
     for (uint j = 0; j < href->len; j++) {
         href->int_scores[j] = href->phred[j] - qual_offset;
-        // cerr << href->int_scores[j] << " ";
     }
-    // cerr << "\n";
 
     const int *p, *q;
     double mean        = 0.0;
@@ -530,6 +564,71 @@ check_quality_scores(RawRead *href, int qual_offset, int score_limit, int len_li
     return 1;
 }
 
+int
+check_poly_g_run(RawRead *href, int offset, int len_limit)
+{
+    //
+    // Search for a run of G nucleotides at the 3' end of the read. In Illumina machines with
+    // two-color chemistry, no color signal is interpreted as a 'G'. However, when sequencing of
+    // a cluster has failed, and no signal is returned, that is also interpreted as a G and
+    // will receive a very high quality score. If we see a run of these Gs at the end of the
+    // sequence, drop the seqeunce.
+    //
+    const int run_limit       = 10;
+    const int polyg_phred_lim = 35;
+    const int exception_lim   = 2;
+    //
+    // Set a stop index to the first element of the sequence, not including any barcode.
+    //
+    int stop = offset;
+
+    //
+    // We will look for a run of a certain length (run_len) while allowing for a small number
+    // of exceptions to the run, either with a different base or a lowerer phred score.
+    int run_len       = 0;
+    int exception_len = 0;
+    
+    //
+    // Set j to be the last element in the sequence.
+    //
+    int j = href->len - 1;
+
+    // cerr << "Start: " << j << "; Stop: " << stop << "\n";
+    
+    while (j >= stop) {
+
+        // cerr << "  Col: " << j << "; Nuc: " << href->seq[j] << "; Phred: " << href->phred[j] << "; Int: " << href->int_scores[j] << "\n";
+        
+        if (href->seq[j] == 'G' && href->int_scores[j] >= polyg_phred_lim)
+            run_len++;
+        else
+            exception_len++;
+
+        if (exception_len >= exception_lim) {
+            // cerr << "Returning 1; run_len: " << run_len << "; exception_len: " << exception_len << "\n";
+            // cerr << "    " << href->seq << "\n    " << href->phred << "\n";
+            return 1;
+        }
+        
+        if (run_len >= run_limit) {
+            // cerr << "Returning -1; run_len: " << run_len << "; exception_len: " << exception_len << "\n";
+            // cerr << "    " << href->seq << "\n    " << href->phred << "\n";
+            if (j < len_limit) {
+                return 0;
+            } else {
+                href->len      = j + 1;
+                href->seq[j]   = '\0';
+                href->phred[j] = '\0';
+                return -1;
+            }
+        }
+
+        j--;
+    }
+
+    return 1;
+}
+
 bool
 correct_barcode(set<string> &bcs, RawRead *href, seqt type, int num_errs)
 {
@@ -619,8 +718,9 @@ filter_adapter_seq(RawRead *href, char *adapter, int adp_len, AdapterHash &adp_k
                    int kmer_size, int distance, int len_limit)
 {
     vector<pair<int, int> > hits;
-    int   num_kmers = href->len - kmer_size + 1;
-    const char *p   = href->seq;
+    int  query_stop = href->inline_bc_len;
+    int   num_kmers = href->len - query_stop - kmer_size + 1;
+    const char *p   = href->seq + query_stop;
     string kmer;
 
     //
@@ -651,13 +751,17 @@ filter_adapter_seq(RawRead *href, char *adapter, int adp_len, AdapterHash &adp_k
 
         // cerr << "Starting comparison at i: "<< i << "; j: " << j << "\n";
 
-        while (i >= 0 && j >= 0) {
+        while (i >= query_stop && j >= 0) {
             if (href->seq[i] != adapter[j])
                 mismatches++;
             i--;
             j--;
         }
-
+        while (j > 0) {
+            mismatches++;
+            j--;
+        }
+        
         if (mismatches > distance)
             continue;
 
