@@ -1,6 +1,6 @@
 // -*-mode:c++; c-style:k&r; c-basic-offset:4;-*-
 //
-// Copyright 2011-2023, Julian Catchen <jcatchen@uoregon.edu>
+// Copyright 2011-2024, Julian Catchen <jcatchen@uoregon.edu>
 //
 // This file is part of Stacks.
 //
@@ -30,14 +30,18 @@
 #include "input.h"
 #include "kmers.h"
 
-enum fastqt   {generic_fastq, illv1_fastq, illv2_fastq};
+enum fastqt   {generic_fastq, illv1_fastq, illv2_fastq, illv3_fastq};
 
 enum barcodet {null_null,
                null_inline,   null_index,
                inline_null,   index_null,
                inline_inline, index_index,
                inline_index,  index_inline};
-enum seqt {single_end, paired_end};
+
+enum seqt     {single_end, paired_end};
+
+enum machinet {genome_analyzer_iix, miseq,
+               hiseq, nextseq, novaseq, miniseq, unknown};
 
 typedef unordered_map<string, vector<int>, std::hash<string> > AdapterHash;
 
@@ -143,34 +147,37 @@ public:
 
 class RawRead {
 public:
-    fastqt fastq_type;
-    char  *inline_bc;
-    char  *index_bc;
-    char  *se_bc;
-    char  *pe_bc;
-    char  *machine;
-    int    run;
-    int    lane;
-    int    tile;
-    int    x;
-    int    y;
-    int    index;
-    int    read;
-    char  *seq;
-    char  *phred;
-    int   *int_scores;
-    bool   filter;
-    int    inline_bc_len;
-    int    retain;
-    uint   size;
-    uint   len;
-    double win_len;
-    double stop_pos;
-
+    fastqt   fastq_type;
+    char    *inline_bc;
+    char    *index_bc;
+    char    *se_bc;
+    char    *pe_bc;
+    char    *_machine;
+    machinet machine_type;
+    char    *umi;
+    int      run;
+    int      lane;
+    int      tile;
+    int      x;
+    int      y;
+    int      index;
+    int      read;
+    char    *seq;
+    char    *phred;
+    int     *int_scores;
+    bool     filter;
+    int      inline_bc_len;
+    int      retain;
+    uint     size;
+    uint     len;
+    double   win_len;
+    double   stop_pos;
+    
     RawRead(uint buf_len, int read, int barcode_size, double win_size) {
         this->inline_bc     = new char[id_len  + 1];
         this->index_bc      = new char[id_len  + 1];
-        this->machine       = new char[id_len  + 1];
+        this->_machine      = new char[id_len  + 1];
+	this->umi           = new char[id_len  + 1];
         this->seq           = new char[buf_len + 1];
         this->phred         = new char[buf_len + 1];
         this->int_scores    = new  int[buf_len];
@@ -189,10 +196,18 @@ public:
 
         this->inline_bc[0] = '\0';
         this->index_bc[0]  = '\0';
-        this->machine[0]   = '\0';
+        this->_machine[0]  = '\0';
+	this->umi[0]       = '\0';
         this->seq[0]       = '\0';
         this->phred[0]     = '\0';
 
+	this->inline_bc[id_len] = '\0';
+        this->index_bc[id_len]  = '\0';
+	this->_machine[id_len]  = '\0';
+	this->umi[id_len]       = '\0';
+
+        this->detect_machine_type();
+        
         this->set_len(buf_len);
 
         this->se_bc = NULL;
@@ -241,7 +256,8 @@ public:
     ~RawRead() {
         delete [] this->inline_bc;
         delete [] this->index_bc;
-        delete [] this->machine;
+        delete [] this->_machine;
+	delete [] this->umi;
         delete [] this->seq;
         delete [] this->phred;
         delete [] this->int_scores;
@@ -259,6 +275,63 @@ public:
 
         return 0;
     }
+    void detect_machine_type() {
+
+        const char *p = this->_machine;
+        switch(*p) {
+        case 'A':
+            if (*(p+1) == 'A')
+                this->machine_type = nextseq;
+            else
+                this->machine_type = novaseq;
+            break;
+        case 'N':
+        case 'V':
+            this->machine_type = nextseq;
+            break;
+        case 'C':
+        case 'D':
+        case 'J':
+        case 'K':
+            this->machine_type = hiseq;
+            break;
+        case 'H':
+            if (*(p+1) != 'W')
+                this->machine_type = novaseq;
+            else if (strncmp(p, "HWI-M", 5) == 0)
+                this->machine_type = miseq;
+            else if (strncmp(p, "HWUSI", 5) == 0)
+                this->machine_type = genome_analyzer_iix;
+            else if (strncmp(p, "HWI-C", 5) == 0)
+                this->machine_type = hiseq;
+            else if (strncmp(p, "HWI-D", 5) == 0)
+                this->machine_type = hiseq;
+            break;
+        case 'M':
+            if (*(p+1) == 'N')
+                this->machine_type = miniseq;
+            else
+                this->machine_type = miseq;
+            break;
+        default:
+            this->machine_type = unknown;
+            break;
+        }
+
+        return;
+    }
+    void machine(const char *p) {
+        strncpy(this->_machine, p, id_len);
+        this->detect_machine_type();
+    }
+    const char *machine() {
+        return this->_machine;
+    }
+    bool check_poly_g() {
+        if (this->machine_type == machinet::novaseq || this->machine_type == machinet::nextseq)
+            return true;
+        return false;
+    }
     int set_len(uint buf_len) {
         if (buf_len == this->len)
             return 0;
@@ -280,7 +353,7 @@ public:
             this->win_len = 1;
 
         this->len     += this->inline_bc_len;
-        this->stop_pos = this->len - this->win_len;
+        this->stop_pos = this->len - this->win_len - 1;
 
         return 0;
     }
@@ -333,6 +406,7 @@ int  filter_adapter_seq(RawRead *, char *, int, AdapterHash &, int, int, int);
 int  init_adapter_seq(int, char *, int &, AdapterHash &);
 
 int  check_quality_scores(RawRead *, int, int, int, int);
+int  check_poly_g_run(RawRead *, int, int);
 
 //
 // Templated function to process barcodes.
