@@ -1,6 +1,6 @@
 // -*-mode:c++; c-style:k&r; c-basic-offset:4;-*-
 //
-// Copyright 2011-2023, Julian Catchen <jcatchen@illinois.edu>
+// Copyright 2011-2024, Julian Catchen <jcatchen@illinois.edu>
 //
 // This file is part of Stacks.
 //
@@ -37,12 +37,14 @@ using namespace std;
 //
 FileT  in_file_type  = FileT::unknown;
 FileT  out_file_type = FileT::unknown;
+InputT in_path_type  = InputT::files;
 string in_file;
 string in_file_p1;
 string in_file_p2;
 string in_path_1;
 string in_path_2;
 string out_path;
+string out_basename;
 string barcode_file;
 string renz_1;
 string renz_2;
@@ -63,6 +65,8 @@ bool     discards        = false;
 bool     overhang        = false;
 bool     filter_illumina = false;
 bool     check_radtag    = true;
+bool     check_poly_g    = false;
+bool     force_poly_g    = false;
 uint     truncate_seq    = 0;
 int      barcode_dist_1  = 1;
 int      barcode_dist_2  = -1;
@@ -147,6 +151,10 @@ int main (int argc, char* argv[]) {
     else
         cerr << "Processing single-end data.\n";
     cerr << "Using Phred+" << qual_offset << " encoding for quality scores.\n";
+    if (force_poly_g)
+        check_poly_g ?
+            cerr << "Checks for poly-G runs have been forced.\n" :
+            cerr << "Checks for poly-G runs have been disabled.\n";
     if (truncate_seq > 0)
         cerr << "Reads will be truncated to " << truncate_seq << "bp\n";
     if (filter_illumina)
@@ -206,7 +214,11 @@ int main (int argc, char* argv[]) {
     int result = 1;
 
     for (uint i = 0; i < files.size(); i++) {
-        cerr << "Processing file " << i+1 << " of " << files.size() << " [" << files[i].first.c_str() << "]\n";
+
+        if (paired)
+            cerr << "Processing file-pair " << i+1 << " of " << files.size() << "\n";
+        else
+            cerr << "Processing file " << i+1 << " of " << files.size() << "\n";
 
 	initialize_counters(counters[files[i].first]);
 
@@ -268,8 +280,10 @@ int main (int argc, char* argv[]) {
         cerr << "-" << counters[files[i].first]["ambiguous"]   << " ambiguous barcodes; "
              << "-" << counters[files[i].first]["noradtag"]    << " ambiguous RAD-Tags; "
              << "+" << counters[files[i].first]["recovered"]   << " recovered; "
-             << "-" << counters[files[i].first]["low_quality"] << " low quality reads; "
-             << counters[files[i].first]["retained"] << " retained reads.\n";
+             << "-" << counters[files[i].first]["low_quality"] << " low quality reads; ";
+        if (check_poly_g)
+            cerr << "-" << counters[files[i].first]["poly_g"]  << " reads with poly-G runs; ";
+        cerr << counters[files[i].first]["retained"] << " retained reads.\n";
         if (filter_adapter)
             cerr << "    "
                  << counters[files[i].first]["adapter"] << " reads with adapter sequence.\n";
@@ -298,7 +312,7 @@ int main (int argc, char* argv[]) {
     }
     cerr << "done.\n";
     
-    print_results(argc, argv, barcodes, counters, barcode_log);
+    print_results(argc, argv, out_basename, barcodes, counters, barcode_log);
 
     cerr << "process_radtags is done.\n";
     
@@ -337,6 +351,7 @@ initialize_counters(map<string, long> &counter)
     counter["ill_filtered"] = 0;
     counter["adapter"]      = 0;
     counter["low_quality"]  = 0;
+    counter["poly_g"]       = 0;
     counter["noradtag"]     = 0;
     counter["ambiguous"]    = 0;
     counter["retained"]     = 0;
@@ -357,6 +372,7 @@ consolidate_counters(map<string, long> &counter,
 	counter["ill_filtered"] += cntr["ill_filtered"];
 	counter["adapter"]      += cntr["adapter"];
 	counter["low_quality"]  += cntr["low_quality"];
+        counter["poly_g"]       += cntr["poly_g"];
 	counter["noradtag"]     += cntr["noradtag"];
 	counter["ambiguous"]    += cntr["ambiguous"];
 	counter["retained"]     += cntr["retained"];
@@ -381,12 +397,14 @@ consolidate_barcode_logs(map<BarcodePair, map<string, long>> &barcode_log,
 		barcode_log[bc]["noradtag"] = 0;
 		barcode_log[bc]["total"]    = 0;
 		barcode_log[bc]["low_qual"] = 0;
+                barcode_log[bc]["poly_g"]   = 0;
 		barcode_log[bc]["retained"] = 0;
 		barcode_log[bc]["proppair"] = 0;
 	    }
 	    barcode_log[bc]["noradtag"] += it->second["noradtag"];
 	    barcode_log[bc]["total"]    += it->second["total"];
 	    barcode_log[bc]["low_qual"] += it->second["low_qual"];
+            barcode_log[bc]["poly_g"]   += it->second["poly_g"];
 	    barcode_log[bc]["retained"] += it->second["retained"];
 	    barcode_log[bc]["proppair"] += it->second["proppair"];
 	}
@@ -473,7 +491,7 @@ process_paired_reads(string prefix_1,
 
     r_1 = new RawRead(strlen(s_1->seq), 1, min_bc_size_1, win_size);
     r_2 = new RawRead(strlen(s_2->seq), 2, min_bc_size_2, win_size);
-
+    
     //
     // Set len_limit so that if we encounter reads already shorter than truncate_seq limit
     // they will be discarded.
@@ -508,6 +526,8 @@ process_paired_reads(string prefix_1,
         parse_input_record(s_2, r_2);
         counter["total"] += 2;
 
+        if (force_poly_g == false && r_1->check_poly_g()) check_poly_g = true;
+        
         //
         // If a BestRAD protocol was used, check which read the barcode/restriction cutsite is on
         // and transpose the reads if necessary.
@@ -748,9 +768,18 @@ process_paired_reads_parallel(string prefix_1,
     }
 
     map<BarcodePair, int> output_thread_map;
-    uint k = 0;
+    //
+    // We must take care because multiple barcodes can point to the same output file,
+    // so we must assign all barcodes for a single output file to the same thread for writing.
+    //
+    map<fhType *, vector<BarcodePair>> rev_fh_map;
     for (auto it = pair_1_fhs.begin(); it != pair_1_fhs.end(); it++) {
-	output_thread_map[it->first] = (k % num_output_threads) + num_input_threads;
+        rev_fh_map[it->second].push_back(it->first);
+    }
+    uint k = 0;
+    for (auto it = rev_fh_map.begin(); it != rev_fh_map.end(); it++) {
+        for (uint j = 0; j < it->second.size(); j++)
+            output_thread_map[it->second[j]] = (k % num_output_threads) + num_input_threads;
 	k++;
     }
     
@@ -998,11 +1027,13 @@ process_paired_reads_parallel(string prefix_1,
 		
 		r_1 = new RawRead(seq_init_len_1, 1, min_bc_size_1, win_size);
 		r_2 = new RawRead(seq_init_len_2, 2, min_bc_size_2, win_size);
-
+    
 		parse_input_record(s_1, r_1);
 		parse_input_record(s_2, r_2);
 		cntr["total"] += 2;
 
+                if (force_poly_g == false && r_1->check_poly_g()) check_poly_g = true;
+                
 		//
 		// If a BestRAD protocol was used, check which read the barcode/restriction cutsite is on
 		// and transpose the reads if necessary.
@@ -1170,7 +1201,7 @@ process_reads(string prefix,
     }
 
     r = new RawRead(strlen(s->seq), 1, min_bc_size_1, win_size);
-
+    
     BarcodePair bc;
     //
     // If no barcodes were specified, set the barcode object to be the input file name so
@@ -1201,6 +1232,8 @@ process_reads(string prefix,
 
         parse_input_record(s, r);
 
+        if (force_poly_g == false && r->check_poly_g()) check_poly_g = true;
+                
         if (barcode_type == inline_null ||
             barcode_type == index_null)
             bc.set(r->se_bc);
@@ -1345,9 +1378,18 @@ process_reads_parallel(string prefix,
     }
 
     map<BarcodePair, int> output_thread_map;
-    uint k = 0;
+    //
+    // We must take care because multiple barcodes can point to the same output file,
+    // so we must assign all barcodes for a single output file to the same thread for writing.
+    //
+    map<fhType *, vector<BarcodePair>> rev_fh_map;
     for (auto it = pair_1_fhs.begin(); it != pair_1_fhs.end(); it++) {
-	output_thread_map[it->first] = (k % num_output_threads) + num_input_threads;
+        rev_fh_map[it->second].push_back(it->first);
+    }
+    uint k = 0;
+    for (auto it = rev_fh_map.begin(); it != rev_fh_map.end(); it++) {
+        for (uint j = 0; j < it->second.size(); j++)
+            output_thread_map[it->second[j]] = (k % num_output_threads) + num_input_threads;
 	k++;
     }
     
@@ -1557,6 +1599,8 @@ process_reads_parallel(string prefix,
 		
 		parse_input_record(s, r);
 
+                if (force_poly_g == false && r->check_poly_g()) check_poly_g = true;
+                
 		bc = new BarcodePair();
 		//
 		// If no barcodes were specified, set the barcode object to be the input file name so
@@ -1708,6 +1752,20 @@ process_singlet(RawRead *href,
         return 0;
     }
 
+    //
+    // Drop this sequence if it has a run of poly-Gs at the 3' end.
+    //   Either the user has forced the check on, and check_poly_g is true, or we turned it on
+    //   internally via href->check_poly_g because of the machine type that generated the data.
+    //
+    if ((check_poly_g || href->check_poly_g()) &&
+        check_poly_g_run(href, href->inline_bc_len, len_limit) <= 0) {
+        counter["poly_g"]++;
+        if (barcode_type != null_null)
+            bc_log["poly_g"]++;
+        href->retain = 0;
+        return 0;
+    }
+
     //
     // Drop this sequence if it contains adapter sequence.
     //
@@ -1896,6 +1954,7 @@ int dist(const char *res_enz, char *seq) {
 
 int
 print_results(int argc, char **argv,
+              string out_basename,
               vector<BarcodePair> &barcodes,
               map<string, map<string, long> > &counters,
               map<BarcodePair, map<string, long> > &barcode_log)
@@ -1903,9 +1962,12 @@ print_results(int argc, char **argv,
     map<string, map<string, long> >::iterator it;
 
     string log_path = out_path + "process_radtags.log";
-    if (!in_path_1.empty()) {
+
+    if (in_path_type == InputT::directory) {
+        //
         // In directory mode, use `$out_path/process_radtags.$(basename $in_path).log`.
         // For consistency we always use realpath().
+        //
         char abspath [PATH_MAX];
         char* rv = realpath(in_path_1.c_str(), abspath);
         if (rv == NULL) {
@@ -1917,6 +1979,13 @@ print_results(int argc, char **argv,
         size_t p = abspath_s.find_last_of('/');
         string in_dir_name = abspath_s.substr(p+1);
         log_path = out_path + "process_radtags." + in_dir_name + ".log";
+        
+    } else if (in_path_type == InputT::files && out_basename.length() > 0) {
+        //
+        // We are executing on a single (-f) or pair of files (-1/-2), and user has
+        // supplied an output basename.
+        //
+        log_path = out_path + out_basename + ".log";
     }
     ofstream log(log_path.c_str());
 
@@ -1943,8 +2012,10 @@ print_results(int argc, char **argv,
         log << "Adapter Seq" << "\t";
     if (bestrad)
 	log << "Transposed Reads" << "\t";
-    log << "Low Quality\t"
-        << "Barcode Not Found\t"
+    log << "Low Quality\t";
+    if (check_poly_g)
+        log << "Poly-G\t";
+    log << "Barcode Not Found\t"
         << "RAD cutsite Not Found\t"
         << "Total\n";
 
@@ -1957,8 +2028,10 @@ print_results(int argc, char **argv,
             log << it->second["adapter"] << "\t";
 	if (bestrad)
 	    log << it->second["transposed"] << "\t";
-        log << it->second["low_quality"] << "\t"
-            << it->second["ambiguous"]   << "\t"
+        log << it->second["low_quality"] << "\t";
+        if (check_poly_g)
+            log << it->second["poly_g"] << "\t";
+        log << it->second["ambiguous"]   << "\t"
             << it->second["noradtag"]    << "\t"
             << it->second["total"]       << "\n";
     }
@@ -1967,6 +2040,7 @@ print_results(int argc, char **argv,
     map<string, long> c;
     c["total"]        = 0;
     c["low_quality"]  = 0;
+    c["poly_g"]       = 0;
     c["adapter"]      = 0;
     c["ill_filtered"] = 0;
     c["ambiguous"]    = 0;
@@ -1981,6 +2055,7 @@ print_results(int argc, char **argv,
         c["total"]        += it->second["total"];
         c["ill_filtered"] += it->second["ill_filtered"];
         c["low_quality"]  += it->second["low_quality"];
+        c["poly_g"]       += it->second["poly_g"];
         c["adapter"]      += it->second["adapter"];
         c["ambiguous"]    += it->second["ambiguous"];
         c["noradtag"]     += it->second["noradtag"];
@@ -2005,6 +2080,7 @@ print_results(int argc, char **argv,
         print_nreads(c["adapter"], "reads contained adapter sequence");
     print_nreads(c["ambiguous"], "barcode not found drops");
     print_nreads(c["low_quality"], "low quality read drops");
+    print_nreads(c["poly_g"], "poly-G run drops");
     print_nreads(c["noradtag"], "RAD cutsite not found drops");
     print_nreads(c["retained"], "retained reads");
 
@@ -2031,8 +2107,10 @@ print_results(int argc, char **argv,
     if (filter_adapter)
         log << "Reads containing adapter sequence\t" << c["adapter"] << "\t" << as_percentage((double) c["adapter"] / c["total"]) << "\n";
     log << "Barcode Not Found\t"     << c["ambiguous"]   << "\t" << as_percentage((double) c["ambiguous"] / c["total"]) << "\n"
-        << "Low Quality\t"           << c["low_quality"] << "\t" << as_percentage((double) c["low_quality"] / c["total"]) << "\n"
-        << "RAD Cutsite Not Found\t" << c["noradtag"]    << "\t" << as_percentage((double) c["noradtag"] / c["total"]) << "\n"
+        << "Low Quality\t"           << c["low_quality"] << "\t" << as_percentage((double) c["low_quality"] / c["total"]) << "\n";
+    if (check_poly_g)
+        log << "Poly-G Runs\t"       << c["poly_g"]      << "\t" << as_percentage((double) c["poly_g"]   / c["total"]) << "\n";
+    log << "RAD Cutsite Not Found\t" << c["noradtag"]    << "\t" << as_percentage((double) c["noradtag"] / c["total"]) << "\n"
         << "Retained Reads\t"        << c["retained"]    << "\t" << as_percentage((double) c["retained"] / c["total"]) << "\n";
     if (paired)    
 	log << "Properly Paired\t"   << c["proppair"]    << "\t" << as_percentage(((double) c["proppair"]*2.0) / c["total"]) << "\n";
@@ -2060,8 +2138,10 @@ print_results(int argc, char **argv,
         log << "Filename\t";
     log << "Total\t"
         << "RAD Cutsite Not Found\t"
-        << "Low Quality\t"
-        << "Retained Reads\t"
+        << "Low Quality\t";
+    if (check_poly_g)
+        log << "Poly-G Runs\t";
+    log << "Retained Reads\t"
 	<< "Pct Retained\t";
     if (paired)
 	log << "Properly Paired\t"
@@ -2078,11 +2158,13 @@ print_results(int argc, char **argv,
             log << barcodes[i].name << "\t";
 
         if (barcode_log.count(barcodes[i]) == 0)
-            log << "0\t" << "0\t" << "0\t" << "0\n";
+            log << "0\t" << "0\t" << "0\t" << "0\t" << "0\n";
         else {
             log << barcode_log[barcodes[i]]["total"]    << "\t"
-                << barcode_log[barcodes[i]]["noradtag"] << "\t"
-                << barcode_log[barcodes[i]]["low_qual"] << "\t"
+                << barcode_log[barcodes[i]]["noradtag"] << "\t";
+            if (check_poly_g)
+                log << barcode_log[barcodes[i]]["poly_g"] << "\t";
+            log << barcode_log[barcodes[i]]["low_qual"] << "\t"
                 << barcode_log[barcodes[i]]["retained"] << "\t"
 		<< as_percentage((double) barcode_log[barcodes[i]]["retained"] / barcode_log[barcodes[i]]["total"]) << "\t";
 	    if (paired)
@@ -2142,6 +2224,8 @@ int parse_command_line(int argc, char* argv[]) {
             {"interleaved",          no_argument, NULL, 'I'},
             {"merge",                no_argument, NULL, 'm'},
             {"disable-rad-check",    no_argument, NULL, 'R'}, {"disable_rad_check",    no_argument, NULL, 'R'},
+            {"disable-poly-g-check", no_argument, NULL, 1004},
+            {"force-poly-g-check",   no_argument, NULL, 1005},
             {"filter-illumina",      no_argument, NULL, 'F'},
             {"retain-header",        no_argument, NULL, 'H'},
             {"bestrad",              no_argument, NULL, 1000},
@@ -2160,6 +2244,7 @@ int parse_command_line(int argc, char* argv[]) {
             {"file",           required_argument, NULL, 'f'},
             {"file-p1",        required_argument, NULL, '1'}, {"file_p1",        required_argument, NULL, '1'},
             {"file-p2",        required_argument, NULL, '2'}, {"file_p2",        required_argument, NULL, '2'},
+            {"basename",       required_argument, NULL, 1003},
             {"in-path",        required_argument, NULL, 'p'},
             {"out-path",       required_argument, NULL, 'o'},
             {"truncate",       required_argument, NULL, 't'},
@@ -2224,8 +2309,9 @@ int parse_command_line(int argc, char* argv[]) {
             in_file = optarg;
             break;
         case 'p':
-            in_path_1 = optarg;
-            in_path_2 = in_path_1;
+            in_path_1  = optarg;
+            in_path_2  = in_path_1;
+            in_path_type = InputT::directory;
             break;
         case '1':
             paired     = true;
@@ -2235,6 +2321,9 @@ int parse_command_line(int argc, char* argv[]) {
             paired     = true;
             in_file_p2 = optarg;
             break;
+        case 1003:
+            out_basename = optarg;
+            break;
         case 'P':
             paired = true;
             break;
@@ -2293,6 +2382,14 @@ int parse_command_line(int argc, char* argv[]) {
         case 'R':
             check_radtag = false;
             break;
+        case 1004:
+            force_poly_g = true;
+            check_poly_g = false;
+            break;
+        case 1005:
+            force_poly_g = true;
+            check_poly_g = true;
+            break;
         case 'F':
             filter_illumina = true;
             break;
@@ -2388,6 +2485,11 @@ int parse_command_line(int argc, char* argv[]) {
     if (in_path_2.length() > 0 && in_path_2.at(in_path_2.length() - 1) != '/')
         in_path_2 += "/";
 
+    if (in_path_1.length() > 0 && out_basename.length() > 0) {
+        cerr << "You can only specify --basename with the -f or -1/-2 options.\n";
+        help();
+    }
+    
     if (out_path.length() == 0)
         out_path = ".";
 
@@ -2477,19 +2579,18 @@ void help() {
          << "  -p,--in-path: path to a directory of files.\n"
          << "  -P,--paired: files contained within the directory are paired.\n"
          << "  -I,--interleaved: specify that the paired-end reads are interleaved in single files.\n"
-         << "  -i,--in-type: input file type, either 'fastq', 'gzfastq' (gzipped fastq), 'bam', or 'bustard' (default: guess, or gzfastq if unable to).\n"
          << "  -b,--barcodes: path to a file containing barcodes for this run, omit to ignore any barcoding or for already demultiplexed data.\n"
-         << "  -o,--out-path: path to output the processed files.\n"
          << "  -f: path to the input file if processing single-end sequences.\n"
          << "  -1: first input file in a set of paired-end sequences.\n"
          << "  -2: second input file in a set of paired-end sequences.\n"
+         << "  -o,--out-path: path to output the processed files.\n"
+         << "    --basename: specify the prefix of the output files when using -f or -1/-2.\n\n"
 	 << "  --threads: number of threads to run.\n"
          << "  -c,--clean: clean data, remove any read with an uncalled base ('N').\n"
          << "  -q,--quality: discard reads with low quality (phred) scores.\n"
          << "  -r,--rescue: rescue barcodes and RAD-Tag cut sites.\n"
          << "  -t,--truncate: truncate final read length to this value.\n"
          << "  -D,--discards: capture discarded reads to a file.\n"
-         << "  -y,--out-type: output type, either 'fastq', 'gzfastq', 'fasta', or 'gzfasta' (default: match input type).\n"
          << "\n"
          << "  Barcode options:\n"
          << "    --inline-null:   barcode is inline with sequence, occurs only on single-end read (default).\n"
@@ -2531,12 +2632,16 @@ void help() {
          << "    --adapter-1 <sequence>: provide adaptor sequence that may occur on the single-end read for filtering.\n"
          << "    --adapter-2 <sequence>: provide adaptor sequence that may occur on the paired-read for filtering.\n"
          << "      --adapter-mm <mismatches>: number of mismatches allowed in the adapter sequence.\n\n"
-         << "  Output options:\n"
+         << "  Input/Output options:\n"
+         << "    --in-type: input file type, either 'fastq', 'gzfastq' (gzipped fastq), 'bam', or 'bustard' (default: guess, or gzfastq if unable to).\n"
+         << "    --out-type: output type, either 'fastq', 'gzfastq', 'fasta', or 'gzfasta' (default: match input type).\n"
          << "    --retain-header: retain unmodified FASTQ headers in the output.\n"
          << "    --merge: if no barcodes are specified, merge all input files into a single output file.\n\n"
          << "  Advanced options:\n"
          << "    --filter-illumina: discard reads that have been marked by Illumina's chastity/purity filter as failing.\n"
          << "    --disable-rad-check: disable checking if the RAD cut site is intact.\n"
+         << "    --force-poly-g-check: force a check for runs of poly-Gs (default: autodetect based on machine type in FASTQ header).\n"
+         << "    --disable-poly-g-check: disable checking for runs of poly-Gs (default: autodetect based on machine type in FASTQ header).\n"
 	 << "    --encoding: specify how quality scores are encoded, 'phred33' (Illumina 1.8+/Sanger, default) or 'phred64' (Illumina 1.3-1.5).\n"
          << "    --window-size: set the size of the sliding window as a fraction of the read length, between 0 and 1 (default 0.15).\n"
          << "    --score-limit: set the phred score limit. If the average score within the sliding window drops below this value, the read is discarded (default 10).\n"
