@@ -31,12 +31,14 @@
 //
 FileT  in_file_type  = FileT::unknown;
 FileT  out_file_type = FileT::unknown;
+InputT in_path_type  = InputT::files;
 string in_file;
 string in_file_p1;
 string in_file_p2;
 string in_path_1;
 string in_path_2;
 string out_path;
+string out_basename;
 string barcode_file;
 char  *adapter_1;
 char  *adapter_2;
